
/*
 * pbrt source code Copyright(c) 1998-2005 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For educational use only; commercial use expressly forbidden.
 * NO WARRANTY, express or implied, for this software.
 * (See file License.txt for complete license)
 */

// plastic.cpp*
#include "pbrt.h"
#include "material.h"
// Anemone Class Declarations
class Anemone : public Material {
public:
	// Anemone Public Methods
  Anemone(Reference<Texture<Spectrum> > kd,
	  Reference<Texture<Spectrum> > ks,
	  Reference<Texture<float> > rough,
	  Reference<Texture<float> > bump,
	  string vertFile,
	  Spectrum sigs,
	  Spectrum siga,
	  float n) {
    Kd = kd;
    Ks = ks;
    roughness = rough;
    bumpMap = bump;
    vFile = vertFile;
    sigS = sigs;
    sigA = siga;
    N = n;

  }
  BSDF *GetBSDF(const DifferentialGeometry &dgGeom,
		const DifferentialGeometry &dgShading) const;
  
  virtual bool ContainsBSSRDF() { return true; }
  
private:
  // Anemone Private Data
  Reference<Texture<Spectrum> > Kd, Ks;
  Reference<Texture<float> > roughness, bumpMap;
  string vFile;
  Spectrum sigS;
  Spectrum sigA;
  float N;  
};
// Anemone Method Definitions
BSDF *Anemone::GetBSDF(const DifferentialGeometry &dgGeom,
		const DifferentialGeometry &dgShading) const {
	// Allocate _BSDF_, possibly doing bump-mapping with _bumpMap_
	DifferentialGeometry dgs;
	if (bumpMap)
		Bump(bumpMap, dgGeom, dgShading, &dgs);
	else
		dgs = dgShading;
	BSDF *bsdf = BSDF_ALLOC(BSDF)(dgs, dgGeom.nn);
	Spectrum kd = Kd->Evaluate(dgs).Clamp();
	BxDF *diff = BSDF_ALLOC(Lambertian)(kd);
	Fresnel *fresnel =
		BSDF_ALLOC(FresnelDielectric)(1.5f, 1.f);
	Spectrum ks = Ks->Evaluate(dgs).Clamp();
	float rough = roughness->Evaluate(dgs);
	BxDF *spec = BSDF_ALLOC(Microfacet)(ks, fresnel,
		BSDF_ALLOC(Blinn)(1.f / rough));
	bsdf->Add(diff);
	bsdf->Add(spec);
	//cfongc
	bsdf->setHasBSSRDF(true);
	bsdf->setVertsFile(vFile);
	bsdf->setSigS(sigS);
	bsdf->setSigA(sigA);
	bsdf->setN(N);
	return bsdf;
}
// Anemone Dynamic Creation Routine
extern "C" DLLEXPORT Material * CreateMaterial(const Transform &xform,
		const TextureParams &mp) {
	Reference<Texture<Spectrum> > Kd = mp.GetSpectrumTexture("Kd", Spectrum(1.f));
	Reference<Texture<Spectrum> > Ks = mp.GetSpectrumTexture("Ks", Spectrum(1.f));
	Reference<Texture<float> > roughness = mp.GetFloatTexture("roughness", .1f);
	Reference<Texture<float> > bumpMap = mp.GetFloatTexture("bumpmap", 2.5);
	string vertsFile = mp.FindString("vertsFile");
	float marbleSigs[] = {2.19,2.62,3.0};
	float marbleSiga[] = {0.0021,0.0041,0.0071};
	Spectrum sigs = mp.FindSpectrum("sigs",Spectrum(marbleSigs));
	Spectrum siga = mp.FindSpectrum("siga",Spectrum(marbleSiga));

	float n = mp.FindFloat("n", 1.5);
	return new Anemone(Kd, Ks, roughness, bumpMap,vertsFile,sigs,siga,n);
}
