
/*
 * pbrt source code Copyright(c) 1998-2005 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For educational use only; commercial use expressly forbidden.
 * NO WARRANTY, express or implied, for this software.
 * (See file License.txt for complete license)
 */

// translucent.cpp*
#include "pbrt.h"
#include "material.h"
// Translucent Class Declarations
class FishScale : public Material {
public:
	// Translucent Public Methods
	FishScale(Reference<Texture<float> > sS,
			Reference<Texture<float> > sA,
			Reference<Texture<float> > G,
			Reference<Texture<float> > n,
			Reference<Texture<Spectrum> > kd,
			Reference<Texture<Spectrum> > kr,
			Reference<Texture<Spectrum> > kt,
			Reference<Texture<float> > bump){				

		sigmaS= sS;
		sigmaA= sA;
		g= G;
		index= n;
		Kd= kd;
		Kr= kr;
		Kt= kt;
		bumpMap= bump;
	}
	BSDF *GetBSDF(const DifferentialGeometry &dgGeom, const DifferentialGeometry &dgShading) const;
private:
	// FishScale Private Data
	Reference<Texture<float> > sigmaS, sigmaA, g, index, bumpMap;
	Reference<Texture<Spectrum> > Kd, Kr, Kt;
};
// FishScale Method Definitions
BSDF *FishScale::GetBSDF(const DifferentialGeometry &dgGeom, const DifferentialGeometry &dgShading) const {
	// Allocate _BSDF_, possibly doing bump-mapping with _bumpMap_
	DifferentialGeometry dgs;
	if (bumpMap)
		Bump(bumpMap, dgGeom, dgShading, &dgs);
	else
		dgs = dgShading;

	BSDF *bsdf = BSDF_ALLOC(BSDF)(dgs, dgGeom.nn);

	Spectrum kd = Kd->Evaluate(dgs).Clamp();
	Spectrum kr = Kr->Evaluate(dgs).Clamp();
	Spectrum kt = Kt->Evaluate(dgs).Clamp();

	float sS= sigmaS->Evaluate(dgs);
	float sA= sigmaA->Evaluate(dgs);
	float phase= g->Evaluate(dgs);

	//lambertian reflection
	if (!kd.Black()) 
		bsdf->Add(BSDF_ALLOC(Lambertian)(kd));

	//specular reflection
	//if (!kr.Black())
	//	bsdf->Add(BSDF_ALLOC(SpecularReflection)(0.1*kr,BSDF_ALLOC(FresnelDielectric)(1.33, index)));

	//glossy surface reflection 
	Fresnel *fresnel = BSDF_ALLOC(FresnelDielectric)(index, 1.33f);
	if (!kr.Black())
			bsdf->Add(BSDF_ALLOC(Microfacet)(kr, fresnel, BSDF_ALLOC(Blinn)(1.f / 0.1f)));
	
	//subsurface reflection
	if (!kr.Black())
		bsdf->Add(BSDF_ALLOC(FirstBRDF)(kr, 1.33, index, sS, sA, phase, 0.01));
	
	//subsurface transmittence
	if (!kt.Black())
		bsdf->Add(BSDF_ALLOC(ZeroBTDF)(kt, 1.33, index, 1.33, 0.01, sS, sA, phase));

	return bsdf;
}
extern "C" DLLEXPORT Material * CreateMaterial(const Transform &xform,
		const TextureParams &mp) {
	Reference<Texture<Spectrum> > kd = mp.GetSpectrumTexture("Kd", Spectrum(1.f));
	Reference<Texture<Spectrum> > kr = mp.GetSpectrumTexture("Kr", Spectrum(1.f));
	Reference<Texture<Spectrum> > kt = mp.GetSpectrumTexture("Kt", Spectrum(1.f));

	Reference<Texture<float> > sigmaS = mp.GetFloatTexture("sigmaS", 50.f);
	Reference<Texture<float> > sigmaA = mp.GetFloatTexture("sigmaA", 3.8f);
	Reference<Texture<float> > g = mp.GetFloatTexture("g", 0.8);
	Reference<Texture<float> > n = mp.GetFloatTexture("n", 1.5);
	Reference<Texture<float> > bumpMap = mp.GetFloatTexture("bumpmap", 0.f);

	return new FishScale(sigmaS, sigmaA, g, n, kd, kr, kt, bumpMap);
}
